__title__ = 'seekr'
__description__ = 'Count small kmer frequencies in nucleotide sequences.'
__url__ = 'https://github.com/CalabreseLab/seekr'
__version__ = '1.5.1'
__author__ = 'Jessime Kirk'
__author_email__ = 'jessime.kirk@gmail.com'
__license__ = 'MIT'
